from pathlib import Path

import typer

from prezmanifest.labeller import LabellerOutputTypes, label

app = typer.Typer(help="Prez Manifest label commands")


@app.command(
    name="iris",
    help="Find all the IRIs of objects in the Manifest's resources without labels",
)
def iris_command(
    manifest: Path = typer.Argument(
        ..., help="The path of the Prez Manifest file to be labelled"
    ),
) -> None:
    for iri in label(manifest, LabellerOutputTypes.iris):
        print(str(iri))


@app.command(
    name="rdf",
    help="Create labels for all the objects in the Manifest's resources without labels",
)
def rdf_command(
    manifest: Path = typer.Argument(
        ..., help="The path of the Prez Manifest file to be labelled"
    ),
    context: str = typer.Argument(
        None,
        help="The path of an RDF file, a directory of RDF files or the URL of a SPARQL endpoint from which t obtain labels",
    ),
) -> None:
    print(
        label(manifest, LabellerOutputTypes.rdf, context).serialize(format="longturtle")
    )


@app.command(
    name="manifest",
    help="Create labels for all the objects in the Manifest's resources without labels and store them as a new Manifest resource",
)
def manifest_command(
    manifest: Path = typer.Argument(
        ..., help="The path of the Prez Manifest file to be labelled"
    ),
    context: Path = typer.Argument(
        ...,
        help="The path of an RDF file, a directory of RDF files or the URL of a SPARQL endpoint from which t obtain labels",
    ),
) -> None:
    label(manifest, additional_context=context)



from pathlib import Path
from textwrap import dedent

import pytest
from rdflib import Graph
from rdflib.compare import isomorphic
from typer.testing import CliRunner

from prezmanifest.cli import app
from prezmanifest.documentor import TableFormats, catalogue, table

runner = CliRunner()


def test_create_table_01():
    expected = dedent(
        """
        Resource | Role | Description
        --- | --- | ---
        Catalogue Definition:<br />[`catalogue.ttl`](catalogue.ttl) | [Catalogue Data](https://prez.dev/ManifestResourceRoles/CatalogueData) | The definition of, and medata for, the container which here is a dcat:Catalog object
        Resource Data:<br />[`vocabs/*.ttl`](vocabs/*.ttl) | [Resource Data](https://prez.dev/ManifestResourceRoles/ResourceData) | skos:ConceptScheme objects in RDF (Turtle) files in the vocabs/ folder
        Profile Definition:<br />[`ogc_records_profile.ttl`](https://github.com/RDFLib/prez/blob/main/prez/reference_data/profiles/ogc_records_profile.ttl) | [Catalogue & Resource Model](https://prez.dev/ManifestResourceRoles/CatalogueAndResourceModel) | The default Prez profile for Records API
        Labels:<br />[`_background/labels.ttl`](_background/labels.ttl) | [Complete Catalogue and Resource Labels](https://prez.dev/ManifestResourceRoles/CompleteCatalogueAndResourceLabels) | An RDF file containing all the labels for the container content                
        """
    ).strip()

    result = table(Path(__file__).parent / "demo-vocabs" / "manifest.ttl")

    print()
    print()
    print(expected)
    print()
    print()
    print(result)
    print()
    print()

    assert result == expected


def test_create_table_02():
    with pytest.raises(ValueError):
        table(Path(__file__).parent / "demo-vocabs" / "manifest-invalid-01.ttl")


def test_create_catalogue():
    expected = Graph().parse(Path(__file__).parent / "demo-vocabs" / "catalogue.ttl")
    actual = catalogue(Path(__file__).parent / "demo-vocabs" / "manifest-cat.ttl")

    assert isomorphic(actual, expected)


def test_create_table_multi():
    expected = dedent(
        """
        Resource | Role | Description
        --- | --- | ---
        Catalogue Definition:<br />[`catalogue-metadata.ttl`](catalogue-metadata.ttl) | [Catalogue Data](https://prez.dev/ManifestResourceRoles/CatalogueData) | The definition of, and medata for, the container which here is a dcat:Catalog object
        Resource Data:<br />[`vocabs/image-test.ttl`](vocabs/image-test.ttl)<br />[`vocabs/language-test.ttl`](vocabs/language-test.ttl) | [Resource Data](https://prez.dev/ManifestResourceRoles/ResourceData) | skos:ConceptScheme objects in RDF (Turtle) files in the vocabs/ folder
        Profile Definition:<br />[`ogc_records_profile.ttl`](https://github.com/RDFLib/prez/blob/main/prez/reference_data/profiles/ogc_records_profile.ttl) | [Catalogue & Resource Model](https://prez.dev/ManifestResourceRoles/CatalogueAndResourceModel) | The default Prez profile for Records API
        Labels:<br />[`_background/labels.ttl`](_background/labels.ttl) | [Complete Catalogue and Resource Labels](https://prez.dev/ManifestResourceRoles/CompleteCatalogueAndResourceLabels) | An RDF file containing all the labels for the container content                
        """
    ).strip()

    result = table(Path(__file__).parent / "demo-vocabs" / "manifest-multi.ttl")

    print()
    print()
    print(expected)
    print()
    print()
    print(result)
    print()
    print()

    assert result == expected


def test_create_table_multi_asciidoc():
    expected = dedent(
        """
        |===
        | Resource | Role | Description
        
        | Catalogue Definition: +
         +
        * link:catalogue-metadata.ttl[`catalogue-metadata.ttl`] | https://prez.dev/ManifestResourceRoles/CatalogueData[Catalogue Data] | The definition of, and medata for, the container which here is a dcat:Catalog object
        | Resource Data: +
         +
        * link:vocabs/image-test.ttl[`vocabs/image-test.ttl`] +
        * link:vocabs/language-test.ttl[`vocabs/language-test.ttl`] | https://prez.dev/ManifestResourceRoles/ResourceData[Resource Data] | skos:ConceptScheme objects in RDF (Turtle) files in the vocabs/ folder
        | Profile Definition: +
         +
        * link:https://github.com/RDFLib/prez/blob/main/prez/reference_data/profiles/ogc_records_profile.ttl[`ogc_records_profile.ttl`] | https://prez.dev/ManifestResourceRoles/CatalogueAndResourceModel[Catalogue & Resource Model] | The default Prez profile for Records API
        | Labels: +
         +
        * link:_background/labels.ttl[`_background/labels.ttl`] | https://prez.dev/ManifestResourceRoles/CompleteCatalogueAndResourceLabels[Complete Catalogue and Resource Labels] | An RDF file containing all the labels for the container content
        |===  
        """
    ).strip()

    result = table(
        Path(__file__).parent / "demo-vocabs" / "manifest-multi.ttl",
        table_format=TableFormats.asciidoc,
    )

    print()
    print()
    print(expected)
    print()
    print()
    print(result)
    print()
    print()

    assert result == expected


def test_create_table_main_entity():
    expected = dedent(
        """
        Resource | Role | Description
        --- | --- | ---
        Catalogue Definition:<br />[`catalogue.ttl`](catalogue.ttl) | [Catalogue Data](https://prez.dev/ManifestResourceRoles/CatalogueData) | The definition of, and medata for, the container which here is a dcat:Catalog object
        Resource Data:<br />[`vocabs/image-test.ttl`](vocabs/image-test.ttl)<br />[`vocabs/language-test.ttl`](vocabs/language-test.ttl) | [Resource Data](https://prez.dev/ManifestResourceRoles/ResourceData) | skos:ConceptScheme objects in RDF (Turtle) files in the vocabs/ folder
        Profile Definition:<br />[`ogc_records_profile.ttl`](https://github.com/RDFLib/prez/blob/main/prez/reference_data/profiles/ogc_records_profile.ttl) | [Catalogue & Resource Model](https://prez.dev/ManifestResourceRoles/CatalogueAndResourceModel) | The default Prez profile for Records API
        Labels:<br />[`_background/labels.ttl`](_background/labels.ttl) | [Complete Catalogue and Resource Labels](https://prez.dev/ManifestResourceRoles/CompleteCatalogueAndResourceLabels) | An RDF file containing all the labels for the container content                
        """
    ).strip()

    result = table(Path(__file__).parent / "demo-vocabs" / "manifest-mainEntity.ttl")

    print()
    print()
    print(expected)
    print()
    print()
    print(result)
    print()
    print()

    assert result == expected


def test_create_catalogue_multi():
    expected = Graph().parse(Path(__file__).parent / "demo-vocabs" / "catalogue.ttl")
    actual = catalogue(Path(__file__).parent / "demo-vocabs" / "manifest-multi.ttl")

    assert isomorphic(actual, expected)


def test_create_catalogue_main_entity():
    expected = Graph().parse(Path(__file__).parent / "demo-vocabs" / "catalogue.ttl")
    actual = catalogue(
        Path(__file__).parent / "demo-vocabs" / "manifest-mainEntity.ttl"
    )

    assert isomorphic(actual, expected)


def test_table_cli():
    result = runner.invoke(
        app,
        [
            "document",
            "table",
            str(Path(__file__).parent / "demo-vocabs" / "manifest.ttl"),
        ],
    )
    actual = result.stdout

    assert "Catalogue Definition" in actual


def test_catalogue_cli():
    expected = Graph().parse(
        data="""
        PREFIX ns1: <http://purl.org/linked-data/registry#>
        PREFIX schema: <https://schema.org/>
        PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
        
        <https://example.com/demo-vocabs>
            a schema:DataCatalog ;
            ns1:status <https://linked.data.gov.au/def/reg-statuses/experimental> ;
            schema:codeRepository "https://github.com/kurrawong/demo-vocabs" ;
            schema:creator <https://kurrawong.ai> ;
            schema:dateCreated "2023"^^xsd:gYear ;
            schema:dateModified "2024-10-16"^^xsd:date ;
            schema:description "A testing catalogue for the Prez Manifest Loader tool" ;
            schema:hasPart
                <https://example.com/demo-vocabs/image-test> ,
                <https://example.com/demo-vocabs/language-test> ;
            schema:name "Demo Vocabularies" ;
            schema:publisher <https://kurrawong.ai> ;
        .        
        """
    )
    result = runner.invoke(
        app,
        [
            "document",
            "catalogue",
            str(Path(__file__).parent / "demo-vocabs" / "manifest.ttl"),
        ],
    )
    actual = Graph().parse(data=result.stdout, format="turtle")

    assert isomorphic(expected, actual)



from pathlib import Path

import typer

from prezmanifest.cli.app import app
from prezmanifest.validator import validate


@app.command(name="validate", help="Validate a Prez Manifest")
def validate_command(
    manifest: Path = typer.Argument(
        ..., help="The path of the Prez Manifest file to be validated"
    ),
) -> None:
    validate(manifest)



"""
Assesses a given Manifest, finds any IRIs in any of the given resources missing labels and tries to patch them from
a given source of labels, such as KurrawongAI's Semantic Background (https://github.com/kurrawong/semantic-background)
repository.
"""

from enum import Enum
from pathlib import Path

from kurra.utils import load_graph
from labelify import extract_labels, find_missing_labels
from rdflib import BNode, Graph, Literal
from rdflib.namespace import PROF

from prezmanifest.definednamespaces import MRR
from prezmanifest.loader import ReturnDatatype, load
from prezmanifest.utils import get_files_from_artifact


class LabellerOutputTypes(str, Enum):
    iris = "iris"
    rdf = "rdf"
    manifest = "manifest"


def label(
    manifest: Path,
    output_type: LabellerOutputTypes = LabellerOutputTypes.manifest,
    additional_context: Path | str | Graph = None,
) -> set | Graph | None:
    """ "Main function for labeller module"""
    # create the target from the Manifest
    manifest_content_graph = load(manifest, return_data_type=ReturnDatatype.graph)

    # determine if any labelling context is given in Manifest
    context_graph = Graph()
    for s, o in manifest_content_graph.subject_objects(PROF.hasResource):
        for role in manifest_content_graph.objects(o, PROF.hasRole):
            if role in [
                MRR.IncompleteCatalogueAndResourceLabels,
                MRR.CompleteCatalogueAndResourceLabels,
            ]:
                for artifact in manifest_content_graph.objects(o, PROF.hasArtifact):
                    artifact: Literal
                    for f in get_files_from_artifact(manifest, artifact):
                        context_graph += load_graph(f)

    # add labels for system IRIs
    context_graph.parse(Path(__file__).parent / "system-labels.ttl")

    if not isinstance(output_type, LabellerOutputTypes):
        raise ValueError(
            f"Invalid output_type value, must be one of {', '.join([x for x in LabellerOutputTypes])}"
        )

    if output_type == LabellerOutputTypes.iris:
        return find_missing_labels(
            manifest_content_graph + context_graph, additional_context
        )

    elif output_type == LabellerOutputTypes.rdf:
        iris = find_missing_labels(manifest_content_graph, context_graph)

        if additional_context is not None:
            return extract_labels(iris, additional_context)
        else:
            return None

    else:  # output_type == LabellerOutputTypes.manifest
        # If this is selected, generate the "rdf" output and create a resource for it in the Manifest
        # If there are no more missing labels then we have a mrr:CompleteCatalogueAndResourceLabels
        # else add mrr:IncompleteCatalogueAndResourceLabels

        # Generate labels for any IRIs missing them, using context given in the Manifest and any
        # Additional Context supplied

        manifest_only_graph = load_graph(manifest)
        rdf_addition = label(manifest, LabellerOutputTypes.rdf, additional_context)

        if len(rdf_addition) > 0:
            new_artifact = manifest.parent / "labels-additional.ttl"
            rdf_addition.serialize(destination=new_artifact, format="longturtle")
            new_resource = BNode()

            # Find the role of any context in the Manifest
            manifest_iri = None
            context_roles = []
            for s, o in manifest_only_graph.subject_objects(PROF.hasResource):
                manifest_iri = s
                for role in manifest_only_graph.objects(o, PROF.hasRole):
                    if role in [
                        MRR.IncompleteCatalogueAndResourceLabels,
                        MRR.CompleteCatalogueAndResourceLabels,
                    ]:
                        context_roles.append(role)

            if (
                MRR.CompleteCatalogueAndResourceLabels in context_roles
                and len(context_roles) == 1
            ):
                # If a CompleteCatalogueAndResourceLabels is present in Manifest and yet more labels were discovered,
                # change CompleteCatalogueAndResourceLabels to IncompleteCatalogueAndResourceLabels and add another
                for s, o in manifest_content_graph.subject_objects(PROF.hasRole):
                    if o == MRR.CompleteCatalogueAndResourceLabels:
                        manifest_only_graph.remove((s, PROF.hasRole, o))
                        manifest_only_graph.add(
                            (manifest_iri, PROF.hasResource, new_resource)
                        )
                        manifest_only_graph.add(
                            (
                                new_resource,
                                PROF.hasRole,
                                MRR.IncompleteCatalogueAndResourceLabels,
                            )
                        )
                        manifest_only_graph.add(
                            (new_resource, PROF.hasArtifact, Literal(new_artifact.name))
                        )
            else:
                # If an IncompleteCatalogueAndResourceLabels was present, add another IncompleteCatalogueAndResourceLabels
                # which together make a CompleteCatalogueAndResourceLabels

                # If none was present, add an IncompleteCatalogueAndResourceLabels or a CompleteCatalogueAndResourceLabels
                # TODO: test for completeness of labelling and add in CompleteCatalogueAndResourceLabels if complete
                manifest_only_graph.add((manifest_iri, PROF.hasResource, new_resource))
                manifest_only_graph.add(
                    (
                        new_resource,
                        PROF.hasRole,
                        MRR.IncompleteCatalogueAndResourceLabels,
                    )
                )
                manifest_only_graph.add(
                    (new_resource, PROF.hasArtifact, Literal(new_artifact.name))
                )

            manifest_only_graph.serialize(destination=manifest, format="longturtle")
        else:
            raise Warning(
                "No new labels have been generated for content in this Manifest. "
                "This could be because none were missing or because no new labels can be found in any "
                "supplied additional context."
            )



from collections.abc import Generator
from pathlib import Path
from typing import List

from kurra.file import load_graph
from rdflib import BNode, Dataset, Graph, Literal, Node, URIRef
from rdflib.namespace import DCAT, OWL, RDF, SDO, SKOS

KNOWN_PROFILES = {
    URIRef("http://www.opengis.net/def/geosparql"): Path(__file__).parent
    / "validator-geosparql-1.1.ttl",
    URIRef("https://data.idnau.org/pid/cp"): Path(__file__).parent
    / "validator-idn-cp.ttl",
    URIRef("https://w3id.org/profile/vocpub"): Path(__file__).parent
    / "validator-vocpub-4.10.ttl",
}

KNOWN_ENTITY_CLASSES = [
    SKOS.ConceptScheme,
    OWL.Ontology,
    DCAT.Resource,
    SDO.CreativeWork,
    SDO.Dataset,
    SDO.DefinedTerm,
]


def get_files_from_artifact(
    manifest_graph: Graph, manifest: Path, artifact: Node
) -> List[Path] | Generator[Path]:
    """Returns an iterable (list or generator) of Path objects for files within an artifact literal.

    This function will correctly interpret artifacts such as 'file.ttl', '*.ttl', '**/*.trig' etc.
    """
    if isinstance(artifact, Literal):
        if "*" not in str(artifact):
            return [manifest.parent / Path(str(artifact))]
        else:
            artifact_str = str(artifact)
            glob_marker_location = artifact_str.find("*")
            glob_parts = [
                artifact_str[:glob_marker_location],
                artifact_str[glob_marker_location:],
            ]

            return Path(manifest.parent / Path(glob_parts[0])).glob(glob_parts[1])
    elif isinstance(artifact, BNode):
        contentLocation = manifest_graph.value(
            subject=artifact, predicate=SDO.contentLocation
        )
        return [manifest.parent / Path(str(contentLocation))]
    else:
        raise TypeError(f"Unsupported artifact type: {type(artifact)}")


def get_identifier_from_file(file: Path) -> List[URIRef]:
    """Returns a list if RDFLib graph identifier (URIRefs) from a triples or quads file
    for all owl:Ontology and skos:ConceptScheme objects"""
    if file.name.endswith(".ttl"):
        g = Graph().parse(file)
        for entity_class in KNOWN_ENTITY_CLASSES:
            v = g.value(predicate=RDF.type, object=entity_class)
            if v is not None:
                return [v]
    elif file.name.endswith(".trig"):
        gs = []
        d = Dataset()
        d.parse(file, format="trig")
        for g in d.graphs():
            gs.append(g.identifier)
        return gs
    else:
        return []


def get_validator(manifest: Path, iri_or_path: URIRef | Literal) -> Graph:
    """Returns a graph from either the path of a SHACL file or a known IRI->profile validator file"""
    if isinstance(iri_or_path, URIRef):
        if iri_or_path not in KNOWN_PROFILES.keys():
            raise ValueError(
                f"You have specified conformance to an unknown profile. Known profiles are {', '.join(KNOWN_PROFILES.keys())}"
            )
        return load_graph(KNOWN_PROFILES[iri_or_path])

    MANIFEST_ROOT_DIR = manifest.parent
    return load_graph(MANIFEST_ROOT_DIR / iri_or_path)
